% Copyright 2008 INRIA, F. Fages & J. Martin
% License GPL
%
% file misc.pl
%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% misc
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% terms storing/retracting
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
store_term(Term, Type) :-
	OldTypeTerm =.. [Type, ListOfTerms],
	call(OldTypeTerm),
	(
	  is_list(Term)
	->
	  append(ListOfTerms, Term, NewListOfTerms)
	;
	  append(ListOfTerms, [Term], NewListOfTerms)
	),
	NewTypeTerm =.. [Type, NewListOfTerms],
	retract(OldTypeTerm),
	assert(NewTypeTerm).

store_term(varorder(Crits)) :-
	cg_mode(dynamic),
	!,
	store_term(Crits, dynvarorder_terms).
store_term(varorder(Crits)) :-
	!,
	store_term(Crits, varorder_terms).
store_term(valorder(Crits)) :-
	cg_mode(dynamic),	
	!,
	store_term(Crits, dynvalorder_terms).
store_term(valorder(Crits)) :-	
	!,
	store_term(Crits, valorder_terms).

reset_dynamic_terms([]).
reset_dynamic_terms([Type | Types]) :-
	Term =.. [Type, _],
	retractall(Term),
	ResetTerm =.. [Type, []],
	assert(ResetTerm),
	reset_dynamic_terms(Types).

% term linearization
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
linearize_term([]) -->
	!.
linearize_term(T) --> 
	{
	 simple(T)
	},
	!,
	[T].
linearize_term(Expr) --> 
	{
	 functor(Expr, F, K),
	 lang(builtins, Builtins),
	 append(Builtins, ['.', ',', ';', '='], Ops),
	 memberchk(F, Ops)
	},
	!,
	linearize_term(0, K, Expr).
linearize_term(Expr) -->
	[F],
	{
	 functor(Expr, F, K)
	},
	linearize_term(0, K, Expr).

linearize_term(K, K, _) -->
	!.
linearize_term(I, K, Expr) -->
	{
	 J is I+1,
	 arg(J, Expr, Arg)
	},
	linearize_term(Arg),
	linearize_term(J, K, Expr).

% term serialization
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
serialize_term('_', '_free_var') :-
	!.
serialize_term(T, ST) :-
	linearize_term(T, LinT, []),
	atomize(LinT, '_', ST).

serialize_int_var_sufx(0, '') :-
	!.
serialize_int_var_sufx(1, '') :-
	!.
serialize_int_var_sufx(I, SI) :-
	serialize_term(I, SI).

serialize_int_var_infx(0, '') :-
	!.
serialize_int_var_infx(I, I).

% counters handling
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
:- dynamic counter/2.

counter_init(Counter) :-
	retractall(counter(Counter,_)).

counter_inc(Counter, N) :-	
	retract(counter(Counter, N)),
	!,
	M is N + 1,
	assert(counter(Counter, M)).
counter_inc(Counter, N) :-
	N is 1,
	assert(counter(Counter, 2)).

%  fresh symbol generation
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
fresh_symbol(Prefix, FreshSymbol) :-
	counter_inc(global_vars, N),
	(
	  N =:= 1
	->
	  FreshSymbol = Prefix
	;
	  number_codes(N, Codes),
	  atom_codes(A, Codes), 
	  atom_concat([Prefix, '_', A], FreshSymbol)
	).

fresh_symbol(FreshSymbol) :-
	counter_inc(global_vars,N),
	number_codes(N, Codes),
	atom_codes(FreshSymbol, [0'S, 0'y, 0'm, 0'_ | Codes]).

% local introduced variable naming
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
fresh_local_var_name(V) :-
	counter_inc(gen_local_var_names, N),
	fresh_local_var_name(N, V).

fresh_local_var_name(N, V) :-
	N < 27,
	!,
	VarCharCode is 65 + N - 1,
	atom_codes(V1, [VarCharCode]),
	atom_concat(['_', V1, '_'], V).
fresh_local_var_name(N, V) :-
	K is 49 + (N // 26) - 1,
	M is N mod 26,	
	VarCharCode is 65 + M,
	atom_codes(V1, [VarCharCode]),
	atom_codes(I, [K]),
	atom_concat(['_', V1, I, '_'], V).

is_var_name(V) :-
	atom(V),
	is_upper_case(V),
	!.	
is_var_name(V) :-
	atom(V),
	atom_codes('_', [USCode]),
	atom_codes(V, [USCode | _]).
	
% atom concatenation
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
atom_concat([A], A) :-
	!.
atom_concat([A,B | T], Con) :-
	atom_concat(A, B, AB),
	atom_concat([AB | T], Con).

atom_concat_sep([A], _S, A) :-
	!.
atom_concat_sep([A,'' | T], S, Con) :-
	!,
	atom_concat_sep([A | T], S, Con).
atom_concat_sep([A,B | T], S, Con) :-
	atom_concat(A, S, AS),
	atom_concat(AS, B, ASB),
	atom_concat_sep([ASB | T], S, Con).

atomize([], _, '') :-
	!.
atomize(['' | T], Sep, At) :-
	!,
	atomize(T, Sep, At).
atomize([H | T], Sep, At) :-
	integer(H),
	!,
	atomize(T, Sep, Asf),
	number_codes(H, As),
	atom_codes(A, As),
	atom_concat_sep([A, Asf], Sep, At).
atomize([H | T], Sep, At) :-
	atomize(T, Sep, Asf),
	atom_concat_sep([H, Asf], Sep, At).

% range to list of integers
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
gen_integers(U, U, [U]) :-
	!.
gen_integers(L, U, [L | Is]) :-
	L < U,
	M is L + 1,
	gen_integers(M, U, Is).

% N fresh variables
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
gen_variables(0, []).
gen_variables(N, [_ | Vs]) :-	
	M is N - 1,
	gen_variables(M, Vs).

% unique int generation from atom
% broken --Mats
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% gen_uid(Atom, Uid) :-
% 	atom_codes(Atom, CharCodes),
% 	atomize(CharCodes,'',At),
% 	atom_codes(At, NCodes),
% 	number_codes(Uid,NCodes).

/*
% List subset
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
subset([],_) :- !.
subset([A | B],C) :-
	member(A,C),
	delete(C,A,C1),
	subset(B,C1).
*/

write_list([]).
write_list([H | T]) :-
	write_list(H),
	write_list(T).
write_list(T) :-
	write(T), nl.

write_list_canonical([]).
write_list_canonical([H | T]) :-	
	write_canonical(H),nl,
	write_list_canonical(T).

integers([]).
integers([I | Is]) :-
	integer(I),
	integers(Is).

varintegers(Is) :-
	varintegers(Is, 0).

varintegers([], N) :-
	N > 0,
	!.
varintegers([I | Is], N) :-
	integer(I),
	M is N + 1,
	!,
	varintegers(Is, M).
varintegers([I | Is], N) :-
	is_var(I),
	varintegers(Is, N).

% characters
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
is_upper_case(Str) :-
	atom_codes(Str, [C | _]),
 	0'A =< C,
	C =< 0'Z.

is_lower_case(Str) :-	
	atom_codes(Str, [C | _]),
	0'a =< C,
	C =< 0'z.

is_underscore(Str) :-
	atom(Str),
	atom_codes(Str, [0'_ | _]).

to_upper_case(Str, StrUpp) :-
	atom_codes(Str, [Ccode | T]),
	0'a =< Ccode,
	Ccode =< 0'z,
	!,
	Cuppcode is Ccode - 32,
	atom_codes(StrUpp, [Cuppcode | T]).
to_upper_case(Str, Str).

to_lower_case(Str, StrLow) :-
	atom_codes(Str, [Ccode | T]),
	0'A =< Ccode,
	Ccode =< 0'Z,
	!,
	Clowcode is Ccode + 32,
	atom_codes(StrLow, [Clowcode | T]).
to_lower_case(Str, Str).

split_string(Str, Sep, StrSplitted) :-
	atom_codes(Str, CharCodes),
	char_code(Sep, SepCharCode),
	split_chars(CharCodes, SepCharCode, SChars, SSplitted),
	(
	  SChars == []
	->
	  StrSplitted = SSplitted
	;
	  atom_codes(StrSChars, SChars),
	  append([StrSChars], SSplitted, StrSplitted)
	).

split_chars([], _, [], []).
split_chars([SepCharCode | CharCodes], SepCharCode, [], Res) :-
	!,
	split_chars(CharCodes, SepCharCode, Splitted, Res0),
	atom_codes(Str, Splitted),
	append([Str], Res0, Res).
split_chars([CharCode | CharCodes], SepCharCode, [CharCode | Splitted], Res) :-
	split_chars(CharCodes, SepCharCode, Splitted, Res).

% split string
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
split(String, Sep) -->
	[Prefix],
	{
	 sub_atom(String, Before, _, After, Sep)
	},
	!,
	{
	 sub_atom(String, 0, Before, _, Prefix),
	 sub_atom(String, _, After, 0, Suffix)
	},
	split(Suffix, Sep).
split(String, _) -->
	[String].

% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
alphanum_concat([], _, []).
alphanum_concat([AlphaNum | T], SepCode, NewCodeList) :-
	alphanum_concat(T, SepCode, CodeListSoFar),
	(
	  integer(AlphaNum)
	->
	  number_codes(AlphaNum, Codes)
	;
	  Codes = AlphaNum
	),
	(
	  CodeListSoFar \== []
	->
	  append(Codes, SepCode, CodeList),
	  append(CodeList, CodeListSoFar, NewCodeList)
	;
	  NewCodeList = Codes
	).

% flips a list
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
flip_list([], []).
flip_list([X], [X]).
flip_list([X1, X2 | Xs], [X2, X1 | Ys]) :-
	flip_list(Xs, Ys).

flip(Term, FTerm) :- 
	is_list(Term),
	!,
	flip_list(Term, FTerm).
flip(Term, FTerm) :-
	compound(Term),
	!,
	Term =.. [Op | Args],
	flip_list(Args, FArgs),
	FTerm =.. [Op | FArgs].
flip(Term, Term).

% function composition
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
composition([F], A, B, R) :- 
	compound(F), 
	F =.. Fs,
	length(Fs, N),
	N =< 2,
	!,
	composition(Fs, A, B, R).
composition([F], A, B, R) :-
	!,
	R =.. [F, A, B].
composition([F | Fs], A, B, R) :-
	!,
	composition(Fs, A, B, R1),
	R =.. [F, R1].
composition(F, A, B, R) :-
	F =.. Fs,
	composition(Fs, A, B, R).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% code generation mode tags {static, dynamic} handling
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
cg_mode(Mode) :-
	code_gen_mode(Mode).

set_cg_mode(Mode) :-
	nonvar(Mode), 
	retractall(code_gen_mode(_)),
	assert(code_gen_mode(Mode)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% search mode tags {direct, reordering} handling
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
srch_mode(Mode) :-
	search_mode(true, Mode).

set_srch_mode(Mode) :-
	nonvar(Mode), 
	retractall(search_mode(_, _)),
	assert(search_mode(true, Mode)).

unset_srch_mode :-
	retractall(search_mode(_, _)),
	assert(search_mode(false, none)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% goal list <-> goal sequence / goal conjunction
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
conj2list(true, []) :-
	!.
conj2list(and(true, B), LB) :-
	!,
	listconj(B, LB).
conj2list(and(A, B), [A | LB]) :-
	listconj(B, LB).

list2conj([], true) :-
	!.
list2conj([H], H) :-
	!.
list2conj([H | T], and(H, CT)) :-
	list2conj(T, CT).

list2seq([], true) :-
	!.
list2seq([H], H) :-
	!.
list2seq([H | T], (H, ST)) :-
	list2seq(T, ST).

first_defunctor([], '') :-
	!.
first_defunctor([H | T], DEF) :-
	is_binder(H),
	!,
	first_defunctor(T, DEF).
first_defunctor([H | T], DEF) :-
	atomic(H),
	functor(Term, H, 1),
	is_logop(Term),
	!,
	first_defunctor(T, DEF).
first_defunctor([H | T], DEF) :-
	atomic(H),
	functor(Term, H, 2),
	is_logop(Term),
	!,
	first_defunctor(T, DEF).
first_defunctor([H | T], DEF) :-
	atomic(H),
	functor(Term, H, 1),
	is_builtinop(Term),
	!,
	first_defunctor(T, DEF).
first_defunctor([H | T], DEF) :-
	atomic(H),
	functor(Term, H, 2),
	is_builtinop(Term),
	!,
	first_defunctor(T, DEF).
first_defunctor([search | T], DEF) :-
	!,
	first_defunctor(T, DEF).
first_defunctor([H | _], H).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% delete uids from records 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
delete_uids(Term, Term) :-
	simple(Term),
	!.
delete_uids(Term, DTerm) :-
	is_list(Term),
	!,
	delete_uids_list(Term, DTerm).
delete_uids(Term, DTerm) :-
	functor(Term, F, K),
	functor(DTerm, F, K),
	delete_uids(0, K, Term, DTerm).

delete_uids(K, K, _, _) :-
	!.
delete_uids(I, K, Term, DTerm) :-
	J is I+1,
	arg(J, Term, X),
	arg(J, DTerm, Y),
	delete_uids(X, Y),
	delete_uids(J, K, Term, DTerm).

delete_uids_list([], []).
delete_uids_list([uid = _ | T], DT) :-
	!,
	delete_uids_list(T, DT).
delete_uids_list([H | T], [H | DT]) :-
	delete_uids_list(T, DT).
