% Copyright 2008 INRIA, F. Fages & J. Martin
% License GPL
%
% file ocg.sp.pl
%

% CLPFD SICStus Prolog operators dictionary 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
dict_sp(<, #<).
dict_sp(=<, #=<).
dict_sp(=, #=).
dict_sp(#, #\=).
dict_sp(>=, #>=).
dict_sp(>, #>).
dict_sp(true,  1).
dict_sp(false, 0).
dict_sp(and, #/\).
dict_sp(or, #\/).
dict_sp(not, #\).
dict_sp(equiv, #<=>).
dict_sp(implies, #=>).
dict_sp(T, T).

dict_sp_dyn(<, fd_lt).
dict_sp_dyn(=<, fd_leq).
dict_sp_dyn(=, fd_eq).
dict_sp_dyn(#, fd_neq).
dict_sp_dyn(>=, fd_geq).
dict_sp_dyn(>, fd_gt).
dict_sp_dyn(true,  1).
dict_sp_dyn(false, 0).
dict_sp_dyn(and, fd_and).
dict_sp_dyn(or, fd_or).
dict_sp_dyn(not, fd_not).
dict_sp_dyn(equiv, fd_equiv).
dict_sp_dyn(implies, fd_implies).
dict_sp_dyn(T, T).

% free variable specialization
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(free_var(X, N, I), Var, [rcp_var(from(Y, N, I), Var)]) :-
	cg_mode(dynamic),
	!,
	delete_uids(X, Y),
	serialize_term(Y, SX),
	to_upper_case(SX, SXUpp),
	serialize_int_var_infx(N, NI),
	serialize_int_var_sufx(I, SI),
	atomize([SXUpp, NI, SI], '_', Var).

% variable specialization
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(var(X, I), STerm, []) :-
	cg_mode(dynamic),
	!,		
	serialize_term(X, SX),
	to_upper_case(SX, SXUpp),
	serialize_int_var_sufx(I, SI),
	atomize([SXUpp, SI], '_', STerm).
generate_sp_code(var(X, I), _STerm, []) :- 
	!,
	raise_exception(code_generation(not_bounded_variable, [var(X, I)], [])).

% prolog {cons, conjunction, disjunction}  
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code('.', '\'.\'', []) :-
	!.

generate_sp_code(',', '\',\'', []) :-
	!.

generate_sp_code(';', '\';\'', []) :-
	!.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% dynamic code generation
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(Term, V, [STerm]) :-
	cg_mode(dynamic),
	srch_mode(_),
	atomic(Term),
	constant_function(Term, _Head, _Body),
	!,
	atomize([Term, search], '_', Functor),
	fresh_local_var_name(V),	
	STerm =.. [Functor, V].


% dynamic atomic (function)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(min_integer, -2^2147483616, []) :-
	!.
generate_sp_code(max_integer, 2^2147483616, []) :-
	!.
generate_sp_code(Term, V, [STerm]) :-
	cg_mode(dynamic),
	atomic(Term),
	constant_function(Term, _Head, _Body),
	!,
	fresh_local_var_name(V),	
	STerm =.. [Term, V].

% dynamic/search atomic (relation)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(Term, V, [STerm]) :-
	cg_mode(dynamic),
	srch_mode(_),
	atomic(Term),
	constant_relation(Term, _Head, _Body),
	!,
	atomize([Term, search], '_', Functor),	
	fresh_local_var_name(V),
	STerm =.. [Functor, V].

% dynamic atomic (relation)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(Term, V, [STerm]) :-
	cg_mode(dynamic),
	atomic(Term),
	constant_relation(Term, _Head, _Body),
	!,
	fresh_local_var_name(V),
	STerm =.. [Term, V].

% dynamic atomic predefined operator
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(Term, STerm, []) :-
	cg_mode(dynamic),
	atomic(Term),
	dict_sp_dyn(Term, STerm),
	!.

% dynamic atomic default
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(Term, Term, []) :-
	cg_mode(dynamic),
	atomic(Term),
	!.

% dynamic length/2 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(length(Term), V, SAdd) :-
	cg_mode(dynamic),
	!,
	generate_sp_code(Term, STerm, SAdd1),
	fresh_local_var_name(V),
	append(SAdd1, [length(STerm, V)], SAdd).

% dynamic nth/2 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(nth(N, Term), V, SAdd) :-
	cg_mode(dynamic),
	!,
	generate_sp_list([N, Term], [SN, STerm], SAdd1),
	fresh_local_var_name(V),
	append(SAdd1, [nth1(SN, STerm, V)], SAdd).

% dynamic pos/2 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(pos(Elem, Term), V, SAdd) :-
	cg_mode(dynamic),
	!,
	generate_sp_list([Elem, Term], [SElem, STerm], SAdd1),
	fresh_local_var_name(V),
	append(SAdd1, [nth1(V, STerm, SElem)], SAdd).

% dynamic/search comparison (<|=<|=|>=|>)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(Term, STerm, SAdd) :-
	cg_mode(dynamic),
	srch_mode(_),
	is_compop(Term),
	!,
	Term =.. [Op | Args],
	dict_sp_dyn(Op, TOp),
	generate_sp_list(Args, SArgs, SAdd),
	STerm =.. [TOp | SArgs].

% dynamic comparison (<|=<|=|>=|>)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(Term, V, SAdd) :-
	cg_mode(dynamic),
	is_compop(Term),
	!,
	Term =.. [Op | Args],
	dict_sp_dyn(Op, TOp),
	generate_sp_list(Args, SArgs1, SAdd1),
	fresh_local_var_name(V),
	append(SArgs1, [V], SArgs),
	STerm =.. [TOp | SArgs],				
	append(SAdd1, [STerm], SAdd).

% dynamic arithmetic exp/2
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(exp(X, E), V, SAdd) :-
	cg_mode(dynamic),
	!,
	generate_sp_list([X, E], [SX, SE], SAdd1),
	fresh_local_var_name(V),
	append(SAdd1, [exp(SX, SE, V)], SAdd).

% dynamic arithmetic log/2
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(log(B, X), V, SAdd) :-
	cg_mode(dynamic),
	!,
	generate_sp_list([B, X], [SB, SX], SAdd1),
	fresh_local_var_name(V),
	append(SAdd1, [log(SB, SX, V)], SAdd).

% dynamic arithmetic (+|-|*|/|min|max|abs)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(Term, V, SAdd) :-
	cg_mode(dynamic),
	is_arithop(Term),
	!,
	Term =.. [Functor | Args],
	generate_sp_list(Args, SArgs, SAdd1),
	fresh_local_var_name(V),
	dict_sp_dyn(Functor, SFunctor),
	STerm =.. [SFunctor | SArgs],
	append(SAdd1, [fd_aeq(STerm, V)], SAdd).

generate_sp_code(';'(P, Q), ';'(P, Q), []) :-
	cg_mode(dynamic),
	srch_mode(_SMode),
	!.
generate_sp_code(','(P, Q), ','(P, Q), []) :-
	cg_mode(dynamic),
	srch_mode(_SMode),
	!.

% dynamic/search not elimination
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(not (P and Q), S, SAdd) :-
	cg_mode(dynamic),
	srch_mode(_SMode),
	!,
	generate_sp_code((not P or not Q), S, SAdd).
generate_sp_code(not (P or Q), S, SAdd) :-
	cg_mode(dynamic),
	srch_mode(_SMode),
	!,
	generate_sp_code((not P and not Q), S, SAdd).
generate_sp_code(not not P, S, SAdd) :-
	cg_mode(dynamic),
	srch_mode(_SMode),
	!,
	generate_sp_code(P, S, SAdd).
generate_sp_code(not P, S, SAdd) :-
	cg_mode(dynamic),
	srch_mode(_SMode),
	is_compop(P),
	!,
	P =.. [F | Args],
	compop_neg(F, NF),
	NP =.. [NF | Args],
	generate_sp_code(NP, S, SAdd).

% dynamic/search disjunction (or)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(P or Q, STerm, SAdd) :-
	cg_mode(dynamic),
	srch_mode(_SMode),
	!,	
	generate_sp_code(P, SP, SAddP),
	generate_sp_code(Q, SQ, SAddQ),
	(
	  is_var_name(SP)
	->
	  append(SAddP, [rcp_fol_call(SP)], SAddP0),
	  list2seq(SAddP0, SPTerm),
	  SAdd0 = []	
	;
	  SPTerm = SP,
	  SAdd0 = SAddP
	),
	(
	  is_var_name(SQ)
	->
	  append(SAddQ, [rcp_fol_call(SQ)], SAddQ0),
	  list2seq(SAddQ0, SQTerm),
	  SAdd = SAdd0	
	;
	  SQTerm = SQ,
	  append(SAdd0, SAddQ, SAdd)
	),
	STerm = ';'(SPTerm, SQTerm).

% dynamic/search conjunction (and)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(P and Q, STerm, SAdd) :-
	cg_mode(dynamic),
	srch_mode(_SMode),	
	!,
	generate_sp_code(P, SP, SAddP),
	generate_sp_code(Q, SQ, SAddQ),
	(
	  is_var_name(SP)
	->
	  append(SAddP, [rcp_fol_call(SP)], SAddP0),
	  list2seq(SAddP0, SPTerm),
	  SAdd0 = []
	;
	  SPTerm = SP,
	  SAdd0 = SAddP
	),
	(
	  is_var_name(SQ)
	->
	  append(SAddQ, [rcp_fol_call(SQ)], SAddQ0),
	  list2seq(SAddQ0, SQTerm),
	  SAdd = SAdd0
	;
	  SQTerm = SQ,
	  append(SAdd0, SAddQ, SAdd)
	),
	STerm = ','(SPTerm, SQTerm).

% dynamic/search implication (implies)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(P implies Q, S, SAdd) :-
	cg_mode(dynamic),
	srch_mode(_SMode),	
	!,	
	generate_sp_code(P, SP, SAddP),
	generate_sp_code(not P, SNP, SAddNP),
	generate_sp_code(Q, SQ, SAddQ),
	list2seq(SAddQ, SAddQSeq),
	fresh_local_var_name(S),
	STerm = (
		  SP,
		  SAddQSeq,
		  fd_equiv(SQ, S, 1)
		;
		  SNP,
		  fd_aeq(S, 1)
		),
	append(SAddP, SAddNP, SAddPNP),
	append(SAddPNP, [STerm], SAdd).

% dynamic conjunction or disjunction (and|or)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(P and Q, ','(SP, SQ), SAdd) :-
	cg_mode(dynamic),
	is_inlined(P),
	is_inlined(Q),
	!,
	generate_sp_code(P, SP, SAddP),
	generate_sp_code(Q, SQ, SAddQ),
	append(SAddP, SAddQ, SAdd).
generate_sp_code(P and Q, SQ, SAdd) :-
	cg_mode(dynamic),
	is_inlined(P),
	!,
	generate_sp_code(P, SP, SAddP1),
	generate_sp_code(Q, SQ, SAddQ),
	append(SAddP1, [SP], SAddP),
	append(SAddP, SAddQ, SAdd).
generate_sp_code(P and Q, SP, SAdd) :-
	cg_mode(dynamic),
	is_inlined(Q),
	!,
	generate_sp_code(P, SP, SAddP),
	generate_sp_code(Q, SQ, SAddQ1),
	append(SAddQ1, [SQ], SAddQ),
	append(SAddP, SAddQ, SAdd).

% dynamic/search logic connectives (not|equiv)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(Term, V, SAdd) :-
	cg_mode(dynamic),		
	is_logop(Term),
	!,
	Term =.. [Functor | Args],
	dict_sp_dyn(Functor, SFunctor),
	generate_sp_list(Args, SArgs1, SAdd1),
	fresh_local_var_name(V),
	append(SArgs1, [V], SArgs),
	STerm =.. [SFunctor | SArgs],
	append(SAdd1, [STerm], SAdd).

% dynamic logic connectives (not|equiv)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(Term, V, [STerm | SAdd1]) :-
	cg_mode(dynamic),	
	is_logop(Term),	
	!,
	Term =.. [Functor | Args],
	dict_sp_dyn(Functor, SFunctor),
	generate_sp_list(Args, SArgs1, SAdd1),
	fresh_local_var_name(V),
	append(SArgs1, [V], SArgs),
	STerm =.. [SFunctor | SArgs].

% dynamic variables/1
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(variables(Term), V, SAdd) :-
	cg_mode(dynamic),	
	!,
	generate_sp_code(Term, STerm, SAdd1),
	fresh_local_var_name(V),
	append(SAdd1, [variables(STerm, V)], SAdd).

% dynamic let/3 (binder)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(let(U, V, E), STerm, SAdd) :-
	cg_mode(dynamic),	
	!,
	cg_mode(dynamic),
	generate_sp_list([U, V], [SU, SV], SAdd1),
	append(SAdd1, [SU = SV], SAdd0),
	generate_sp_code(E, STerm, SAdd2),
	append(SAdd0, SAdd2, SAdd).

% labeling
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(labeling(Term), rcp_labeling(STerm), SAdd) :-
	cg_mode(dynamic),	
	!,
	generate_sp_code(Term, STerm, SAdd).

% dynamic optim. (minimize/1)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(minimize(CostTerm), true, SAdd) :-
 	cg_mode(dynamic),
 	!,	
 	generate_sp_code(CostTerm, SCostTerm, SAddCost),
 	STerm = minimize(rcp_labeling([SCostTerm]), SCostTerm),
 	append(SAddCost, [STerm], SAdd).

% dynamic optim. (minimize/2)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(minimize(SearchTerm, CostTerm), true, SAdd) :-
 	cg_mode(dynamic),
 	!,	
 	generate_sp_code(SearchTerm, SSearchTerm1, SAddSearch1),	
	(	  
	  is_var_name(SSearchTerm1)
	->
	  append([rcp_fol_call(SSearchTerm1)], SAddSearch1, SAdd0),
	  list2seq(SAdd0, SSearchTerm)
	;	  
	  SAdd0 = SAddSearch1,
	  list2seq(SAdd0, STerm0),
	  SSearchTerm = (STerm0, SSearchTerm1)
	),
 	generate_sp_code(CostTerm, SCostTerm, SAddCost),
 	STerm = minimize((SSearchTerm, rcp_labeling([SCostTerm])) ,SCostTerm),
 	append(SAddCost, [STerm], SAdd).

% dynamic optim. (maximize/1)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(maximize(CostTerm), true, SAdd) :-
 	cg_mode(dynamic),
 	!,	
 	generate_sp_code(CostTerm, SCostTerm, SAddCost),
 	STerm = maximize(rcp_labeling([SCostTerm]), SCostTerm),
 	append(SAddCost, [STerm], SAdd).

% dynamic optim. (maximize)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(maximize(SearchTerm, CostTerm), true, SAdd) :-
 	cg_mode(dynamic),
 	!,	
 	generate_sp_code(SearchTerm, SSearchTerm1, SAddSearch1),	
	(	  
	  is_var_name(SSearchTerm1)
	->
	  append([rcp_fol_call(SSearchTerm1)], SAddSearch1, SAdd0),
	  list2seq(SAdd0, SSearchTerm)
	;	  
	  SAdd0 = SAddSearch1,
	  list2seq(SAdd0, STerm0),
	  SSearchTerm = (STerm0, SSearchTerm1)
	),
 	generate_sp_code(CostTerm, SCostTerm, SAddCost),
 	STerm = minimize((SSearchTerm, rcp_labeling([SCostTerm])) ,SCostTerm),
 	append(SAddCost, [STerm], SAdd).

% dynamic attribute
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(fot(Functor, [Arg]), V, SAdd) :-
	cg_mode(dynamic),
	!,
	generate_sp_code(Arg, SArg, SAdd1),
	fresh_local_var_name(V),
 	append(SAdd1, [rcp_att(Functor, SArg, V)], SAdd).

generate_sp_code(att(Term), V, SAdd) :-
	cg_mode(dynamic),
	!,
	generate_sp_code(Term, STerm, SAdd1),
	STerm =.. [Functor, SArg],
	fresh_local_var_name(V),
 	append(SAdd1, [rcp_att(Functor, SArg, V)], SAdd).

% dynamic range
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code([L .. U], V, SAdd) :-
	cg_mode(dynamic),
	!,
	generate_sp_list([L, U], [SL, SU], SAdd1),
	fresh_local_var_name(V),
 	append(SAdd1, [rcp_range([SL..SU], V)], SAdd).

% dynamic "inplace" cons
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(cons(HTerm, TTerm), [SHTerm | STTerm], SAdd) :-
	cg_mode(dynamic),
	!,
	generate_sp_code(HTerm, SHTerm, SAdd1),
	generate_sp_code(TTerm, STTerm, SAdd2),
	append(SAdd1, SAdd2, SAdd).

% dynamic "inplace" list
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(list(Term), STerm, SAdd) :-
	cg_mode(dynamic),
	!,
	generate_sp_list(Term, STerm, SAdd).

% dynamic list
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(Term, STerm, SAdd) :-
	cg_mode(dynamic),
	is_list(Term),
	!,
	generate_sp_list(Term, STerm, SAdd).

% dynamic record
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(rcp_rec(Term), V, SAdd) :-
	cg_mode(dynamic),
	!,
	generate_sp_record(Term, STerm, SAdd1),
	fresh_local_var_name(V),
 	append(SAdd1, [rcp_rec(STerm, V)], SAdd).

% dynamic search 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(search(Term), STerm, SAdd) :-
	cg_mode(dynamic),
	!,
	set_srch_mode(reordering),
	generate_sp_code(Term, STerm, SAdd),
	unset_srch_mode.

% dynamic {variable, value}_ordering/1
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(Term, STerm, SAdd) :-
	cg_mode(dynamic),
	
	functor(Term, Functor, Arity),
	lang(vordercrit_preds, OrderPreds),
	memberchk(Functor/Arity, OrderPreds),
	!,
	Term =.. [Functor | Args],
	generate_sp_code(list(Args), SArgs, SAdd),
	STerm =.. [Functor | SArgs].

% dynamic {greatest, least}/2
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(Term, STerm, SAdd) :-
	cg_mode(dynamic),
	functor(Term, Functor, Arity),
	lang(order_preds, OrderPreds),
	memberchk(Functor/Arity, OrderPreds),
	!,
	Term =.. [Functor | [Args]],
	generate_sp_code(list(Args), SArgs, SAdd),
	STerm =.. [Functor | [SArgs]].

% dynamic {variable, value}_order_criterion/2
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(Term, STerm, SAdd) :-
	cg_mode(dynamic),
	Term =.. [Functor | Args],
	length(Args, Arity),
	lang(dynvordercrit_preds, Preds),
	memberchk(Functor/Arity, Preds),
	!,
	generate_sp_list(Args, SArgs, SAdd),
	STerm =.. [Functor | SArgs].

% dynamic labeling 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(Term, STerm, SAdd) :-
	cg_mode(dynamic),
	
	Term =.. [Functor | Args],
	length(Args, Arity),
	lang(label_preds, LabelPreds),
	memberchk(Functor/Arity, LabelPreds),
	!,
	generate_sp_list(Args, SArgs, SAdd),
	STerm =.. [Functor | SArgs].

% dynamic global constraint all_different
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(all_different(Variables), all_different(SVariables), SAdd) :-
	cg_mode(dynamic),
	!,
	generate_sp_code(Variables, SVariables, SAdd).

% dynamic domain
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(Term, STerm, SAdd) :-
	cg_mode(dynamic),
	Term =.. [Functor | Args],
	is_domain(Term),
	!,
	generate_sp_list(Args, SArgs, SAdd),
	STerm =.. [Functor | SArgs].

% dynamic domain lower bound
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(domain_min(Term), V, SAdd) :-
	cg_mode(dynamic),
	!,
	generate_sp_code(Term, STerm, SAdd1),
	fresh_local_var_name(V),
 	append(SAdd1, [fd_min(STerm, V)], SAdd).

% dynamic domain upper bound
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(domain_max(Term), V, SAdd) :-
	cg_mode(dynamic),
	!,
	generate_sp_code(Term, STerm, SAdd1),
	fresh_local_var_name(V),
 	append(SAdd1, [fd_max(STerm, V)], SAdd).

% dynamic domain cardinality
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(domain_size(Term), V, SAdd) :-
	cg_mode(dynamic),
	
	!,
	generate_sp_code(Term, STerm, SAdd1),
	fresh_local_var_name(V),
 	append(SAdd1, [fd_size(STerm, V)], SAdd).

% dynamic dynamic (idempotence)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(dynamic(Term), STerm, SAdd) :-
	cg_mode(dynamic),
	!,	
	generate_sp_code(Term, STerm, SAdd).

% dynamic/search default 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(Term, V, SAdd) :-
	cg_mode(dynamic),
	srch_mode(_),
	!,	
	Term =.. [Op | Args],
	generate_sp_list(Args, SArgs, SAdd1),
	dict_sp_dyn(Op, TOp),
	(
	  (
	    is_builtinop(Op)
	  ;
	    is_inlined(Term)
	  )
	->
	  Functor = TOp
	;	  
	  atomize([TOp, search], '_', Functor)
	),
	fresh_local_var_name(V),
	append(SArgs, [V], VSArgs),
	STerm =.. [Functor | VSArgs],
	append(SAdd1, [STerm], SAdd).

% % dynamic "inplace" default
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% generate_sp_code(term(Term), STerm, SAdd) :-
% 	cg_mode(dynamic),
% 	
% 	!,	
% 	Term =.. [Op | Args],
% 	generate_sp_list(Args, SArgs, SAdd),
% 	dict(Op, TOp),
% 	STerm =.. [TOp | SArgs].

% % dynamic var/val ordering term 
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% generate_sp_code(Term, STerm, SAdd) :-
% 	cg_mode(dynamic),
% 	
% 	subterm(^, Term),
% 	!,
% 	Term =.. [F | Args],
% 	generate_sp_code(list(Args), SArgs, SAdd),
% 	STerm =.. [F | SArgs].

generate_sp_code(','(P, Q), ','(P, Q), []) :-
	cg_mode(dynamic),
	!.
generate_sp_code(';'(P, Q), ';'(P, Q), []) :-
	cg_mode(dynamic),
	!.
generate_sp_code('.'(P, Q), '.'(P, Q), []) :-
	cg_mode(dynamic),
	!.

% dynamic term default
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(Term, V, SAdd) :-
	cg_mode(dynamic),
	!,
	Term =.. [Op | Args],
	generate_sp_list(Args, SArgs, SAdd1),
	fresh_local_var_name(V),
	append(SArgs, [V], VSArgs),
	dict_sp_dyn(Op, TOp),
	STerm =.. [TOp | VSArgs],
	append(SAdd1, [STerm], SAdd).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% static terms specialization
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% atom 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(Term, STerm, []) :-
	
	atom(Term),
	atom_codes(Term, [0'$ | T]),
	!,
	atom_codes(STerm, T).

% variable 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(free_var(X, N, I), Var, [rcp_var(from(Y, N, I), Var)]) :- 	
	!,
	delete_uids(X, Y),
	serialize_term(Y, SX),
	to_upper_case(SX, SXUpp),
	serialize_int_var_infx(N, NI),
	serialize_int_var_sufx(I, SI),
	atomize([SXUpp, NI, SI], '_', Var).

% dynamic tag
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(dynamic(Term), STerm, []) :-
	!,
	set_cg_mode(dynamic),
	generate_sp_code(Term, STerm1, SAdd1),
	(	  
	  is_var_name(STerm1)
	->
	  append([rcp_fol_call(STerm1)], SAdd1, SAdd),
	  list2seq(SAdd, STerm)
	;	  
	  list2seq(SAdd1, STerm0),
	  STerm = (STerm0, STerm1)
	).

% search
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(search(Formula), SFormula, SAdd) :-
	!,	
	subst_and_or(Formula, FormulaSubst),	
	generate_sp_code(FormulaSubst, SFormula, SAdd).

% labeling
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(labeling(Vars), STerm, []) :-
	!,
	order_labelings([labeling(Vars)], OTerm),
	generate_sp_code(OTerm, STerm, _).

% optimization
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(Term, STerm, SAdd) :-
	functor(Term, Functor, Arity),
	lang(optim_preds, OptimPreds),
	memberchk(Functor/Arity, OptimPreds),
	!,	
	(
	  Arity=:=2
	->
	  Term =.. [Functor,SearchTree,Cost]
	;
	  (
	    Arity=:=1
	  ->
	    Term =.. [Functor,Cost],
	    SearchTree = none
	  ;
	    raise_exception(code_generation(not_defined, [Term], []))
	  )
	),
	variables_set(Cost, CostVariables),
	generate_sp_code(CostVariables, SCostVariables, _),
	generate_sp_code(Cost, SCost, _),
	(
	  Functor == maximize
	->
	  CostLabelOpts = [down]
	;
	  (
	    Functor == minimize
	  ->
	    CostLabelOpts = [up]
	  ;
	    raise_exception(code_generation(not_defined, [Term], []))
	  )
	),
	CostLabeling = labeling(CostLabelOpts, SCostVariables),	
	(
	  SearchTree == none
	->
	  BBGoal = (CostLabeling),
	  SAddST = []
	;
	  generate_sp_code(SearchTree, SSearchTree, SAddST),
	  BBGoal = (SSearchTree, CostLabeling)
	),
	generate_sp_code(BBGoal, SBBGoal, SAddG),
	append(SAddST, SAddG, SAdd),
	STerm =.. [Functor, SBBGoal, SCost].

% fdvar. domain constraint
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(X in L, SX in SL, SAdd) :-	
	!,
	generate_sp_code(X, SX, _),
	generate_sp_range_list(L, SL, SAdd).

% fdvar. domain constraint
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(domain(VarsTerm, Min, Max), domain(SVars, VMin, VMax), SAdd) :- 	
	!,
	variables(VarsTerm, Vars),
	generate_sp_list(Vars, SVars, SAdd1),
	generate_sp_list([Min, Max], [SMin, SMax], SAdd2),
	(
	  integer(SMin)
	->
	  VMin = SMin,
	  SAdd3 = SAdd1
	;
	  fresh_local_var_name(VMin),
	  append(SAdd1, [VMin #= SMin], SAdd3)
	),
	(
	  integer(SMax)
	->
	  VMax = SMax,
	  SAdd4 = SAdd3
	;	  
	  fresh_local_var_name(VMax),
	  append(SAdd3, [VMax #= SMax], SAdd4)
	),
	append(SAdd4, SAdd2, SAdd).

% fdvar. domain lower bound
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(domain_min(Var), FDMin, [fd_min(SVar, FDMin)]) :-
	!,
	generate_sp_code(Var, SVar, _).

% fdvar. domain upper bound
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(domain_max(Var), FDMax, [fd_max(SVar, FDMax)]) :-
	!,
	generate_sp_code(Var, SVar, _).

% fdvar. domain cardinality
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(domain_size(Var), FDMax, [fd_size(SVar, FDMax)]) :-
	!,
	generate_sp_code(Var, SVar, _).

% global constraint disjoint2
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(disjoint2(Xs, Ys, Ss), disjoint2(Data), []) :-
	!,
	generate_sp_code([Xs,Ys,Ss], SArgs, _),
	append(SArgs, [Data], DisjDataArgs),
	DisjDataTerm =.. [disjoint_data | DisjDataArgs],
	call(DisjDataTerm).

% global constraint cumulative
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(cumulative(Os, Ds, Hs, L), STerm, SAdd) :-
	!,	
	generate_sp_code([Os, Ds, Hs], [SOs, SDs, SHs], _),
	set_cg_mode(dynamic),
	generate_sp_code(L, SL, SAdd),
	set_cg_mode(static),
	length(Os, N),
	gen_variables(N, Vs),
	append([SOs, SDs, SHs, Vs], [Tasks], MktaskTermArgs),
	MktaskTerm =.. [mktasks | MktaskTermArgs],
	call(MktaskTerm),
	STerm = (
		  domain(Vs, 0, 1000),
		  cumulative(Tasks, [limit(SL), global(true)])
		).

% global constraint geost with rules
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(geost_constraint(PreObjects, PreSboxes, Dims),
		 geost(Objects, Sboxes, [non_overlapping(objects(Oids), D)]),
		 SAdd) :-
	object_language(clpfdsp),
	!,
	generate_sp_geosr_object(PreObjects, ObjectsX, Oids, ObjectsShapeDoms),	
	generate_sp_code(ObjectsX, Objects, _),
	generate_sp_geosr_sbox(PreSboxes, Sboxs),
	generate_sp_list(ObjectsShapeDoms, SAdd, _),
	gen_integers(1,Dims, D),
	remove_dups(Sboxs, Sboxes).

% gui terms
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(gui_name(Var, AlphaNumList), 
		 gui_varnames([Var], [Atom]), []) :-
	!,
	atom_codes('_', SepCode),
	alphanum_concat(AlphaNumList, SepCode, NameCodeList),
	CL1 = [39 | NameCodeList],
	append(CL1,[39], CL),
	atom_codes(Atom, CL).

generate_sp_code(gui_control(button, Name, Term), 
		 gui_button(BTerm, STerm), SAdd) :-
	object_language(clpfdguisp),
	generate_sp_code(Term, STerm, SAdd),
	!,
	variables_set(STerm, Variables),
	BTerm =..[Name | Variables].

% atomic 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(Term, STerm, []) :-
	atomic(Term),
	dict_sp(Term, STerm),
	!.
generate_sp_code(Term, Term, []) :-
	atomic(Term),
	!.

% list 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(Term, STerm, SAdd) :-
	is_list(Term),
	!,
	generate_sp_list(Term, STerm, SAdd).

% arithmetic 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(Term, STerm, SAdd) :-
	is_arithop(Term),
	Term =.. [Op | Args],
	contains_compop(Args),
	!,
	reify_compop(Args, RArgs, SAdd),
	ST =.. [Op | RArgs],
	eval(ST, STerm).

% search (label) node
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(node(_L, Term), STerm, SAdd) :-
	!,
	generate_sp_code(Term, STerm, SAdd).

% user output 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(write(Term), STerm, SAdd) :-	
	!,
	generate_sp_code(Term, STerm, SAdd),
	format("\"--- User Output ---\"\n ~w ", [Term]),
	format("\n \"~~>\"\n",[]),
	format(" ~w\n\"--- End User Output ---\"\n", [STerm]).

generate_sp_code(A = B, STerm, SAdd) :-
	!,
	generate_sp_code(A, SA, SAddA),
	generate_sp_code(B, SB, SAddB),
	append(SAddA, SAddB, SAdd),	
	(
	  is_builtin_expr(and(A, B))
	->
	  dict_sp(=, TOp),
	  ST =.. [TOp, SA, SB],
	  eval(ST, STerm)
	;
	  STerm =.. [=, SA, SB]
	).

generate_sp_code(equiv(A, B), STerm, SAdd) :-
	!,
	generate_sp_code(A, SA, SAddA),
	generate_sp_code(B, SB, SAddB),
	append(SAddA, SAddB, SAdd),
	(
	  is_builtin_expr(and(A, B))
	->	    
	  dict_sp(equiv, TOp),
	  ST =.. [TOp, SA, SB],
	  eval(ST, STerm)
	;
	  STerm =.. [=, SA, SB]
	).

% default specialization
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(Term, STerm, SAdd) :-
	Term =.. [Op | Args],
	generate_sp_list(Args, SArgs, SAdd),
	dict_sp(Op, TOp),
	ST =.. [TOp | SArgs],
	eval(ST, STerm).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% datastructure-dedicated code generation
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% list 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_list([], [], []).
generate_sp_list([UT | UTerms], [ST | STerms], SAdd) :-
	generate_sp_code(UT, ST, SA),
	generate_sp_list(UTerms, STerms, SAs),
	append(SA, SAs, SAdd).

% record 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_record([], [], []).
generate_sp_record([Key = Val | T], [Key = SVal | ST], SAdd) :-
	generate_sp_code(Val, SVal, SA),
	generate_sp_record(T, ST, SAs),
	append(SA, SAs, SAdd).

% range list 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_range_list([L .. U], (L .. U), []) :-
	integer(L),
	integer(U),
	!.
generate_sp_range_list([L .. U], (inf .. U), []) :-
	L == min_integer,
	integer(U),
	!.
generate_sp_range_list([L .. U], (L .. sup), []) :-
	integer(L),
	U == max_integer,
	!.
generate_sp_range_list([S], {S}, []) :- 
	!,
	integer(S).
generate_sp_range_list([L .. U | T], RUnion \/ (L..U), SAdd) :-
	!,
	integer(L),
	integer(U),
	generate_sp_range_list(T, RUnion, SAdd).
generate_sp_range_list([S | T], RUnion \/ {S}, SAdd) :-
	integer(S),
	generate_sp_range_list(T, RUnion, SAdd).



% geosr  
%%%%%%%%%%%%%%%%%%%%
generate_sp_geosr_object([], [], [], []).
generate_sp_geosr_object([PreObj | PreObjs], [Obj | Objs], [Oid | Oids], ObjShpDoms) :-
	generate_sp_geosr_object(PreObjs, Objs, Oids, OSDoms),	
	PreObj = [Oid, S, Sids, Origins],
	
	(
	 (
	   is_var(S)
	 ;
	   var(S)
	 )
	->
	  fresh_symbol('S', Sid),
	  append([Sid in Sids, element(S, Sids, Sid)], OSDoms, ObjShpDoms)
	;
	  nth1(S, Sids, Sid),
	  ObjShpDoms = OSDoms
	),
	Obj =.. [object, Oid, Sid, Origins, []].

generate_sp_geosr_sbox([], []).
generate_sp_geosr_sbox([PreSbox | PreSboxes], Sboxes) :-
	generate_sp_geosr_sbox(PreSboxes, Sboxs),
	traverse_presbox_shapes(PreSbox, Sbox),
	append(Sbox, Sboxs, Sboxes).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% statement object code generation
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% foldr instance
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_statement(Head, Body1, foldr, SHead-SBody) :-
	!,
				%write('----------'),nl,
				%write(Head),nl,
				%write(Body1),nl,
				%write('----------'),nl,
	Body1 = [RecCall, Body],
	Head =.. [FHead | HeadArgs],
	HeadArgs = [_,_, O, _],
	generate_sp_list(HeadArgs, SHeadArgs1, _SA),
	SHeadArgs1 = [[H | T], VL, _, I],
	SHeadArgs = [[H | T], VL],
	SHead1 =.. [FHead | SHeadArgs],	
	counter_init(gen_local_var_names),
	(
	  srch_mode(_)
	->	  
	  atomize([FHead, search], '_', SFHead)	
	;
	  subst_free_var(Body, SHead1, BodySFV),
	  partition_term([BodySFV]),
	  SFHead = FHead
	),
	cstr_terms(Constraints),
	generate_sp_code(list(Constraints), SBody1, SAdd1),
	delete_variables(SBody1, SBodyWOV, LV),		
	(
	  LV == novar
	->
	  (
	    (
	      SBodyWOV = [X],
	      X \== H
	    )
	  ->
	    V = X
	  ;
	    V = H
	  ),
	  SAdd1 = SBody2
	;
	  append(SAdd1, SBodyWOV, SBody2),
	  V = LV
	),	
	generate_sp_code(RecCall, VRecCall, SAddRecCall),
	CTerm =.. [O, V, VRecCall],
	generate_sp_code(CTerm, SCTerm, SCTermAdd),
	append(SBody2, SCTermAdd, SBody3),	
	append(SAddRecCall, SBody3, SBody),	
	append(SHeadArgs, [SCTerm], RHeadArgs),
	SHead =.. [SFHead | RHeadArgs],		
	TCHead =.. [SFHead, [], '_', I],
	assert(statement(TCHead, ['!'], foldr)).

% foldl instance
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% generate_sp_statement(Head, Body1, foldl, SHead-SBody) :-
% 	srch_mode(_),
% 	
% 	!,
% 		 		write('----------'),nl,
%  				write(Head),nl,
%  				write(Body1),nl,
%  				write('----------'),nl,
% 	Body1 = [Body, RecCall],
% 	Head =.. [FHead | HeadArgs],
% 	HeadArgs = [cons(HX, TX), list(VLX), O],
% 	generate_sp_list([cons(HX, TX), var('I', O), list(VLX)], SHeadArgs, _SA),
% 	SHeadArgs = [[H | _T], I, _VL],
% 	SHead1 =.. [FHead | SHeadArgs],	
% 	counter_init(gen_local_var_names),
% 	(
% 	  srch_mode(_)
% 	->	  
% 	  atomize([FHead, search], '_', SFHead)	
% 	;
% 	  subst_free_var(Body, SHead1, BodySFV),
% 	  partition_term([BodySFV]),
% 	  SFHead = FHead
% 	),	
% 	cstr_terms(Constraints),
% 	generate_sp_code(list(Constraints), SBody1, SAdd1),	
% 	write(Constraints),nl,
% 	delete_variables(SBody1, SBodyWOV, LV),		
% 	(
% 	  LV == novar
% 	->
% 	  (
% 	    (
% 	      SBodyWOV = [X],
% 	      X \== H
% 	    )
% 	  ->
% 	    V = X
% 	  ;
% 	    V = H
% 	  )	  
% 	;
% 	  append(SAdd1, SBodyWOV, SBody2),
% 	  V = LV
% 	),	
% 	generate_sp_code(RecCall, VRecCall, SAddRecCall),
% 	fresh_local_var_name(RetVar),
% 	append(SBody2, SAddRecCall, SBody0),
% 	append(SBody0, [VRecCall, fd_aeq(RetVar, 1)], SBody),
% 	append(SHeadArgs, [RetVar], RHeadArgs),
% 	SHead =.. [SFHead | RHeadArgs],
% 				write('----------'),nl,
% 				write(SHead),nl,
% 				write(SBody),nl,nl,nl,	
% 	TCHead =.. [SFHead, [], I,  '_', I],	
% 	assert(statement(TCHead, ['!'], foldr)).

generate_sp_statement(Head, Body1, foldl, SHead-SBody) :-
	!,
%  				write('----------'),nl,
%  				write(Head),nl,
%  				write(Body1),nl,
%  				write('----------'),nl,
	Body1 = [Body, RecCall],
	Head =.. [FHead | HeadArgs],
	HeadArgs = [cons(HX, TX), list(VLX), O],
	generate_sp_list([cons(HX, TX), var('I', O), list(VLX)], SHeadArgs, _SA),
	SHeadArgs = [[H | _T], I, _VL],
	SHead1 =.. [FHead | SHeadArgs],	
	counter_init(gen_local_var_names),
	(
	  srch_mode(_)
	->	  
	  atomize([FHead, search], '_', SFHead)	
	;
	  subst_free_var(Body, SHead1, BodySFV),
	  partition_term([BodySFV]),
	  SFHead = FHead
	),	
	cstr_terms(Constraints),
	generate_sp_code(list(Constraints), SBody1, SAdd1),
	delete_variables(SBody1, SBodyWOV, LV),
% 	write(sadd1(SAdd1)),nl,
% 	write(sbody1(SBody1)),nl,
	(
	  LV == novar
	->
	  (
	    (
	      SBodyWOV = [X],
	      X \== H
	    )
	  ->
	    V = X
	  ;
	    V = H
	  ),
	  SAdd1 = SBody2
	;
	  append(SAdd1, SBodyWOV, SBody2),
	  V = LV
	),
% 	write(sbody2(SBody2)),nl,	
	generate_sp_code(RecCall, VRecCall, SAddRecCall),
	composition(O, I, V, CTerm),
% 	write(composition(O, I, V, CTerm)),nl,
	generate_sp_code(CTerm, SCTerm1, SCTermAdd1),	
	(
	  is_var_name(SCTerm1)
	->
	  SCTerm = SCTerm1,
	  SCTermAdd = SCTermAdd1
	;
	  fresh_local_var_name(W),
	  SCTerm = W,
	  append(SCTermAdd1, [SCTerm1], SCTermAdd)
	),
	last(SAddRecCall, RecCall1),
	RecCall1 =.. [Functor, Tail, EnvList, FV],
	RecCall2 =.. [Functor, Tail, SCTerm, EnvList, FV],								
	delete(SAddRecCall, RecCall1, SAddRecCall1),	
	append(SAddRecCall1, [RecCall2], SAddRecCall2),	
	append(SBody2, SCTermAdd, SBody3),
	append(SBody3, SAddRecCall2, SBody),
	append(SHeadArgs, [VRecCall], RHeadArgs),
	SHead =.. [SFHead | RHeadArgs],
% 				write('----------'),nl,
% 				write(SHead),nl,
% 				write(SBody),nl,nl,
	TCHead =.. [SFHead, [], I,  '_', I],	
	assert(statement(TCHead, ['!'], foldr)).

% definition
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_statement(Head, _Body, _T, none-none) :-
	srch_mode(_),
	functor(Head, Functor, Arity),
	lang(dynvordercrit_preds, VOrderPreds),	
	memberchk(Functor/Arity, VOrderPreds),
	!.
generate_sp_statement(Head, Body, _T, SHead-SBody) :-
	!,
%  				write('----------'),nl,
%  				write(Head),nl,
%  				write(Body),nl,
%  				write('----------'),nl,
	Head =.. [FHead | HeadArgs],
	generate_sp_list(HeadArgs, SHeadArgs, _SA),
	SHead1 =.. [FHead | SHeadArgs],
	counter_init(gen_local_var_names),	
	(
	  srch_mode(_)
	->	  
	  atomize([FHead, search], '_', SFHead)	
	;
	  subst_free_var(Body, SHead1, BodySFV),
	  partition_term([BodySFV]),
	  SFHead = FHead
	),
	cstr_terms(Constraints),
	generate_sp_code(list(Constraints), SBody1, SAdd1),	
	delete_variables(SBody1, SBodyWOV, LV),
% 				write(body(SBody1)),nl,
% 				write(add(SAdd1)),nl,nl,
	(
	  LV == novar
	->
	  (
	    SBody1 = [Sbody1Content],
	    is_list(Sbody1Content)
	  ->
	    fresh_local_var_name(RArgHead),
	    delete(SBodyWOV, Sbody1Content, SBodyWOV0),
	    append(SBodyWOV0, [RArgHead = Sbody1Content], AppSBody1),
	    append(SAdd1, AppSBody1, SBody)
	  ;
	    fresh_local_var_name(RArgHead),
	    append(SBodyWOV, [fd_aeq(RArgHead, 1)], AppSBody1),
	    append(SAdd1, AppSBody1, SBody)
	  )
	;
	  RArgHead = LV,
	  append(SAdd1, SBodyWOV, SBody)
	),	
	append(SHeadArgs, [RArgHead], RHeadArgs),
	SHead =.. [SFHead | RHeadArgs].

generate_sp_statement(Head, Body, T, _-_) :-
	Term =.. [T, Head, Body],
	object_language(ObjLang),
	raise_exception(code_generation(not_accepted_by_language(ObjLang),
					 [Term], [])).

generate_sp_statements :-
	findall(statement(Head, Body, T),
		statement(Head, Body, T),
		LStatements),
	cg_mode(CGModeBefore),
	generate_sp_statements(LStatements),
	set_cg_mode(CGModeBefore).

generate_sp_statements([]) :-
	!.
generate_sp_statements([statement(Head, Body, T) | LDynClauses]) :-
				%write(statement(Head, Body, T)),nl,
	init_objcodegenerator,	
	set_cg_mode(dynamic),
	retract(statement(Head, Body, T)),	
	generate_sp_statement(Head, Body, T, SHead-SBody),
	assert_statement(SHead, SBody, T),
	
 	set_srch_mode(reordering),
 	counter_init(gen_local_var_names),
 	generate_sp_statement(Head, Body, T, SHeadSrch-SBodySrch),
 	unset_srch_mode,
	assert_statement(SHeadSrch, SBodySrch, T),
	generate_sp_statements(LDynClauses).

assert_statement(none, none, _Type) :-
	!.
assert_statement(Head, Body, Type) :-
	assert(statement(Head, Body, Type)).
