% Copyright 2008 INRIA, F. Fages & J. Martin
% License GPL
%
% file parser.pl
%

:- ensure_loaded([lang]).

:- dynamic constant/2, function/2, rule/2, query/1.
:- dynamic code_gen_mode/1.
:- dynamic conjorder_terms/1, disjorder_terms/1.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% parser 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parse(Files) :-
	init_parser,
	parse_files(Files).

init_parser :-
	retractall(rule(_,_)),
	retractall(constant(_,_)),
	retractall(function(_,_)),
	retractall(query(_)),	
	init_rcp_path,
				%retractall(current_dir(_)),
				%assert(current_dir('./')),	
	retractall(conjorder_terms(_)),
	assert(conjorder_terms([])),
	retractall(disjorder_terms(_)),
	assert(disjorder_terms([])).

parse_files([]).
parse_files([File | Files]) :-		
	parse_file(File),
	parse_files(Files).

parse_file(File) :-
	format("Opening file \"~w\"... \n", [File]),	
	open(File, read, Stream),
	format("Reading file \"~w\"... \n", [File]),	
	parse_stream(Stream),
	format("File \"~w\" read.\n", [File]),	
	close(Stream),
	format("File  \"~w\" closed. \n\n", [File]),
	!.

parse_stream(Stream) :-
	at_end_of_stream(Stream),
	!.
parse_stream(Stream) :-
	read_term(Stream, Term, [variable_names(VarNames), syntax_errors(error)]),
	name_variables(Term, VarNames, VNTerm),
	parse_term(VNTerm),	
	parse_stream(Stream).

parse_term(end_of_file) :-
	!.
parse_term(Term) :-
	parse_statement(Term).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% var 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parse_var(var(Term, I), var(Term, I)) :-
	!.
parse_var(free_var(Term, A, I), free_var(Term, A, I)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% bool
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parse_bool(Term, Term) :-
	is_bool(Term).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% int
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parse_int(Term, Term) :-
	integer(Term).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% infinite
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parse_inf(inf, inf) :-
	!.
parse_inf(sup, sup).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% varint
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parse_varint(Term, PTerm) :-
	parse_var(Term, PTerm),
	!.
parse_varint(Term, PTerm) :-
	parse_int(Term, PTerm). 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% varbool
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parse_varbool(Term, PTerm) :-
	parse_var(Term, PTerm),
	!.
parse_varbool(Term, PTerm) :-
	parse_bool(Term, PTerm).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% ident
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parse_ident(Term, Term) :-
	is_ident(Term).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% name
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parse_name(Term, PTerm) :-
	parse_ident(Term, PTerm),
	!.
parse_name(Name:Ident, PName:PIdent) :-
	parse_name(Name, PName),
	parse_ident(Ident, PIdent).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% criterion
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parse_fcrit(C, fot(Functor, PArgs)) :-
	C =.. [Functor | Args],
	length(Args, Arity),
	lang(folordercrit_preds, Preds),
	memberchk(Functor/Arity, Preds),
	!,
	parse_expr(Args, PArgs).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% head
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% ident[(var, ..., var)]
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parse_head(Term, PTerm) :-
	parse_ident(Term, PTerm),
	!.
parse_head(Term, PTerm) :-
	compound(Term),
	\+ is_list(Term),
	Term =.. [F | Args],	
	parse_ident(F, PF),
	parse_list_var(Args, PArgs),
	PTerm =.. [PF | PArgs].

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% fol
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% varbool
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parse_fol(Term, PTerm) :- 
	parse_varbool(Term, PTerm),
	!.

% in 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parse_fol((V in E), (PV in PE)) :-
	!,
	parse_expr(V, PV),
	parse_expr(E, PE).

% let
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parse_fol(let(V, E, F), PTerm) :-
	parse_fol(quant(let, [V, E, F]), PTerm).

% forall
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parse_fol(forall(V, E, F), PTerm) :-
	parse_fol(quant(forall, [V, E, F]), PTerm).

% exists
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parse_fol(exists(V, E, F), PTerm) :-
	parse_fol(quant(exists, [V, E, F]), PTerm).

% fol quantifiers (let | exists | forall)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parse_fol(quant(Q, [X, E, F]), PTerm) :-
	!,
	parse_or_excpt(var, X, V),
	parse_or_excpt(expr, E, L),
	parse_or_excpt(fol, F, Fol),
	PTerm =.. [Q, V, L, Fol].

% fold right (fol)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parse_fol(foldr(X, E, O, N, F), foldr(V, L, O, Id, Fol)) :-
	!,
	parse_or_excpt(var, X, V),
	parse_or_excpt(expr, E, L),
	parse_or_excpt(expr, F, Fol),
	parse_or_excpt(fol, N, Id).

% fold left (fol)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parse_fol(foldl(X, E, O, N, F), foldl(V, L, O, Id, Fol)) :-
	!,
	parse_or_excpt(var, X, V),
	parse_or_excpt(expr, E, L),
	parse_or_excpt(expr, F, Fol),
	parse_or_excpt(fol, N, Id).

% fol connectives 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parse_fol((not forall(X, E, F)), PTerm) :-
	!,
	parse_fol(exists(X, E, (not F)), PTerm).

parse_fol((not exists(X, E, F)), PTerm) :-
	!,
	parse_fol(forall(X, E, (not F)), PTerm).

parse_fol((not P), (not PP)) :-
	!,
	parse_fol(P, PP).

parse_fol((P and Q), (PP and PQ)) :-
	!,
	parse_fol(P, PP),
	parse_fol(Q, PQ).

parse_fol((P or Q), (PP or PQ)) :-
	!,
	parse_fol(P, PP),
	parse_fol(Q, PQ).

parse_fol((P implies Q), (PP implies PQ)) :-
	!,
	parse_fol(P, PP),
	parse_fol(Q, PQ).

parse_fol((P equiv Q), (PP equiv PQ)) :-
	!,
	parse_fol(P, PP),
	parse_fol(Q, PQ).

% name
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parse_fol(Term, zot(PTerm)) :-
	parse_name(Term, PTerm),
	!.

% (expr compop expr)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parse_fol(Term, PTerm) :-
	compound(Term),
	is_compop(Term),
	!,
	Term =.. [Op, ExprL, ExprR],
	parse_expr(ExprL, PExprL),
	parse_expr(ExprR, PExprR),
	PTerm =.. [Op, PExprL, PExprR].

% fol ordering (conjunct | disjunct)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parse_fol(Term, true) :-
	Term =.. [Functor, CritList],	
	lang(folorder_preds, Preds),
	memberchk(Functor/1, Preds),
	is_list(CritList),
	!,	
	parse_list_fcrit(CritList, PCritList),	
	lang(conjorder_preds, CPreds),
	(
	  memberchk(Functor/1, CPreds)
	->
	  FTerm = conjorder_terms,
	  DynFTerm = dynconjorder_terms
	;
	  FTerm = disjorder_terms,
	  DynFTerm = dyndisjorder_terms
	),	
	(
	  cg_mode(dynamic)
	->
	  store_term(PCritList, DynFTerm)
	;
	  store_term(PCritList, FTerm)
	).

% {minimize, maximize}(Goal, Cost)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parse_fol(minimize(search(Term), Cost), fot(minimize, PArgs)) :-
	!,	
	parse_list([search(Term), Cost], PArgs).
parse_fol(minimize(Term, Cost), fot(minimize, PArgs)) :-
	!,	
	parse_list([search(Term), Cost], PArgs).

parse_fol(maximize(search(Term), Cost), fot(maximize, PArgs)) :-
	!,	
	parse_list([search(Term), Cost], PArgs).
parse_fol(maximize(Term, Cost), fot(maximize, PArgs)) :-
	!,	
	parse_list([search(Term), Cost], PArgs).

% name(expr, ..., expr)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parse_fol(Term, fot(PFunctor,PArgs)) :-
	compound(Term),
	Term =.. [Functor | Args],
	parse_name(Functor, PFunctor),
	parse_list(Args, PArgs).

% exception
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% parse_fol(Term, _) :-
% 	raise_exception(parsing(not_parsed_as(expr), [Term], [])).	


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% expr
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% varint 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parse_expr(Term, PTerm) :-
	parse_varint(Term, PTerm),
	!.

% special atoms
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parse_expr(^, ^) :-
	!.

% infinite
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parse_expr(Term, PTerm) :-
	parse_inf(Term, PTerm),
	!.

% range
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parse_expr([L..U], [PL..PU]) :-
	!,
	parse_expr(L, PL),
	parse_expr(U, PU).

% list 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parse_expr(Term, PTerm) :- 
	is_list(Term),
	!,
	parse_list(Term, PTerm).

% record 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parse_expr(Term, rcp_rec(PTerm)) :- 
	parse_record(Term, PTerm),
	!.

parse_expr(rcp_rec(Term), rcp_rec(PTerm)) :-
	!,
	parse_record(Term, PTerm).

% conditional 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parse_expr((Expr if TExpr), (PExpr if PTExpr)) :-
	!,
	parse_expr(Expr, PExpr),
	parse_test(TExpr, PTExpr).

% let
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parse_expr(let(X, E, F), let(V, L, Expr)) :-
	!,
	parse_or_excpt(var, X, V),
	parse_or_excpt(expr, E, L),
	parse_or_excpt(expr, F, Expr).

% map 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parse_expr(map(X, E, F), PTerm) :- 
	parse_expr(foldr(X, E, cons, [], F), PTerm),
	!.

% fold right (expr)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parse_expr(foldr(X, E, O, N, F), foldr(V, L, O, Id, Expr)) :-
	!,
	parse_or_excpt(var, X, V),
	parse_or_excpt(expr, E, L),
	parse_or_excpt(expr, F, Expr),
	parse_or_excpt(expr, N, Id).

% fold left (expr)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parse_expr(foldl(X, E, O, N, F), foldl(V, L, O, Id, Expr)) :-
	!,
	parse_or_excpt(var, X, V),
	parse_or_excpt(expr, E, L),
	parse_or_excpt(expr, F, Expr),
	parse_or_excpt(expr, N, Id).

% filter
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parse_expr(filter(P,E), filter(PP, PE)) :-
	!,
	parse_or_excpt(name, P, PP),
	parse_or_excpt(expr, E, PE).
	
% expr arithop expr 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parse_expr(Term, PTerm) :-
	is_arithop(Term),
	!,
	Term =.. [Op, ExprL, ExprR],
	parse_expr(ExprL, PExprL),
	parse_expr(ExprR, PExprR),
	PTerm =.. [Op, PExprL, PExprR].

% arithop(expr)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parse_expr(Term, PTerm) :-
	is_arithop(Term),
	!,	
	Term =.. [Op, Expr],
	parse_expr(Expr, PExpr),
	PTerm =.. [Op, PExpr].

% fol
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parse_expr(Term, PTerm) :-
	parse_fol(Term, PTerm).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% list of ordered pairs 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parse_list([], []).
parse_list([(K = V) | Ts], [(PK = PV) | PTs]) :-
	parse_name(K, PK),
	parse_expr(V, PV),	
	parse_list(Ts, PTs).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% list 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parse_list([], []).
parse_list([T | Ts], [PT | PTs]) :-
	parse_expr(T, PT),
	parse_list(Ts, PTs).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% list of variables
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parse_list_var([], []).
parse_list_var([T | Ts], [PT | PTs]) :-
	parse_var(T, PT),
	parse_list_var(Ts, PTs).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% list of criteria (on conjuncts or disjuncts)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parse_list_fcrit([], []).
parse_list_fcrit([C | Cs], [PC | PCs]) :-
	parse_fcrit(C, PC),
	parse_list_fcrit(Cs, PCs).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% record 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parse_record(Term, PTerm) :-
	Term =.. [{}, (OrderedPairSeq)],
	parse_orderedpair_seq((OrderedPairSeq), PTerm).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% sequence of ordered pairs 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parse_orderedpair_seq((K = V), [(PK = PV)]) :-
	!,
	parse_name(K, PK),
	parse_expr(V, PV).
parse_orderedpair_seq(Term, [(PK = PV) | PTs]) :-
	Term =.. [',', (K = V), Comps],	
	parse_name(K, PK),
	parse_expr(V, PV),	
	parse_orderedpair_seq(Comps, PTs).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% test
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parse_test((LExpr is RExpr), is(LExpr, RExpr)) :-
	parse_expr(LExpr, _PLExpr),
	parse_expr(RExpr, _PRExpr).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% statements
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parse_statement((Head --> Fol)) :-
        parse_head(Head, PHead),
	parse_fol(Fol, PFol),
	assert(rule(PHead, PFol)),
	!.

parse_statement((Head = Expr)) :-
	functor(Head, _, 0),
	name_anonymous_variables(Expr),	
	parse_head(Head, PHead),
	parse_expr(Expr, PExpr),
	variables(PHead, BindedVars),
	name_free_variables(PExpr, BindedVars, PExprOLV),
	assert(constant(PHead, PExprOLV)),
	!.

parse_statement((Head = Expr)) :-
	name_anonymous_variables(Expr),
	parse_head(Head, PHead),
	parse_expr(Expr, PExpr),
	variables(PHead, BindedVars),
	name_free_variables(PExpr, BindedVars, PExprOLV),
	assert(function(PHead, PExprOLV)),
	!.

parse_statement((? Fol)) :-
	parse_fol(Fol, PFol),
	assert(query(PFol)),
	!.

parse_statement(import(FileName)) :- 
	atom(FileName),
	atom_concat([FileName, '.rcp'], F),
	rcp_path(Root),	
	atom_concat([Root, F], File),
	parse_file(File),
	!.

parse_statement(Term) :-
	raise_exception(parsing(not_parsed_as(statement), [Term], [])).	

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% parse or raise exception
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% int
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parse_or_excpt(int, Term, PTerm) :-
	!,
	parse_int(Term, PTerm).

% var
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parse_or_excpt(var, Term, PTerm) :-
	!,
	parse_var(Term, PTerm).

% name
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parse_or_excpt(name, Term, PTerm) :-
	!,
	parse_name(Term, PTerm).

% fol
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parse_or_excpt(fol, Term, PTerm) :-
	!,
	parse_fol(Term, PTerm).

% expr
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
parse_or_excpt(expr, Term, PTerm) :-
	!,
	parse_expr(Term, PTerm).

parse_or_excpt(Type, Term, _PTerm) :-
	raise_exception(parsing(not_parsed_as(Type), [Term], [])).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% "type" checking 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% is Term a boolean ?
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
is_bool(Term) :- is_true(Term).
is_bool(Term) :- is_false(Term).

is_true(true).
is_true(1).

is_false(false).
is_false(0).

% is Term a variable ?
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
is_free_var(free_var(_, _, _)).

is_var(var(_, _)).
is_var(free_var(_, _, _)).

% is Term an anonymous variable ?
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
is_anonym_var(Term) :-
	is_free_var(Term),
	arg(1, Term, Name),	
	is_underscore(Name).

% is Term an identifier ?
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
is_ident(Term) :-
	atom(Term),
	is_lower_case(Term).

% is Term a range ?
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
is_range([_ .. _]).

% is Term an arithmetic operator ?
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
is_arithop(Term) :-
	functor(Term, Functor, Arity),
	lang(arith_ops, Ops),
	memberchk(Functor/Arity, Ops).

% is Term a comparison operator ?
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
is_compop(Term) :-
	functor(Term, Functor, Arity),
	lang(comp_ops, CompOps),
	memberchk(Functor/Arity, CompOps).

% is Term a logic operator ?
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
is_logop(Term) :-
	functor(Term, Functor, Arity),
	lang(logic_ops, LogOps),
	memberchk(Functor/Arity, LogOps).

% is Term a list operator ?
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
is_listop(Term) :-
	functor(Term, Functor, Arity),
	lang(list_ops, Ops),
	memberchk(Functor/Arity, Ops).

% is Term an arithmetic operator ?
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
is_builtinop(Term) :-
	functor(Term, Functor, Arity),		
	lang(builtins, Ops),
	memberchk(Functor/Arity, Ops).

% is Term an optim. pred ?
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
is_optim(Term) :-
	functor(Term, Functor, Arity),
	lang(optim_preds, Preds),
	memberchk(Functor/Arity, Preds).

% is Term a variable binder ?
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
is_binder(Term) :-
	lang(bind_preds, BindPreds),
	memberchk(Term, BindPreds).

% is Term inlined ?
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
is_inlined(Term) :-
	functor(Term, Functor, Arity),	
	lang(inlined_preds, Preds),
	memberchk(Functor/Arity, Preds).

% utilities
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
init_rcp_path :-
	rcp_path(_),
	!.
init_rcp_path :-	
	environ('_', PWD),
	split_string(PWD, '/', StrSplitted),
	write(StrSplitted),nl,
	sublist(StrSplitted, Part, 0, _, 4),	
	atomize(Part, '/', Dir),
	atom_concat(['/', Dir, '/'], Directory),
	assert(rcp_path(Directory)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% includes
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- ensure_loaded([misc]).
