:- module(r2cp, [init_interpreter/0,
		 rcp_var/2,
		 rcp_anavar/2,
		 rcp_att/3,
		 rcp_rec/2,
		 rcp_range/2,
		 rcp_call/2,
		 rcp_fol_call/1,
		 variables/2,
		 rcp_variables/2,
		 variables_set/2,
		 rcp_variables_set/2,
		 max/3, min/3,
		 log/3, exp/3,
		 fd_lt/3, fd_lt/2,
		 fd_leq/3, fd_leq/2,
		 fd_geq/3, fd_geq/2,
		 fd_gt/3, fd_gt/2,
		 fd_neq/3, fd_gt/2,
		 fd_aeq/2,
		 fd_and/3,
		 fd_or/3,
		 fd_implies/3,
		 fd_equiv/3,
		 geost/3,
		 variable_ordering/1,
		 value_ordering/1,
		 rcp_var_ordering/1,
		 rcp_val_ordering/1,
		 rcp_labeling/1,
		 rcp_rec_uid/2]).

:- use_module(library(lists)).
:- use_module(library(terms)).
:- use_module(library(avl)).
:- use_module(library(sets)).
:- use_module(library(clpfd)).
:- use_module(library(atts)).
:- use_module(library(samsort)).

:- use_module(['geosr/geosr.pl']).

:- dynamic rcp_rec_uid/2.
:- dynamic rcp_var_ordering/1,
	   rcp_val_ordering/1.

:- attribute rcp_var_name/1.
verify_attributes(_, _, []).

% statement
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
:-op(900, fx, ?).

% such that 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
:- op(800, xfx, if).

% logics
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
:- op(780, xfx, implies).
:- op(780, xfx, equiv).
:- op(775, yfx, or).
:- op(770, yfx, and).
:- op(710,  fy, not).

% comparison
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
:- op(700, xfx, #).

% domain
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
:- op(700, xfx, in).

% range
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
:- op(550, xfx, ..).

init_interpreter :-
	init_rcp_rec_uid,
	retractall(rcp_var_ordering(_)),
	assert(rcp_var_ordering([])),
	retractall(rcp_val_ordering(_)),
	assert(rcp_val_ordering([])).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% r2cp-specific constructs handling
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% variables (global name-variable linking)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
rcp_var(Name, Value) :-	
	execution_state(break_level(0), private(Priv)),
	memberchk(r2cp(Priv0), Priv),
	memberchk((Name = Value), Priv0),
	(
 	  var(Value)
 	->
 	  put_atts(Value, rcp_var_name(Name))
 	;
 	  true
 	).
	
% variables 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
rcp_anavar(Value, Name) :-
	get_atts(Value, rcp_var_name(Name)).

% attribute (accessor)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
rcp_att(Att, rcp_rec(AssocList), Value) :-
	memberchk(Att = Value, AssocList).

% record (uid handling)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
rcp_rec(AssocList, rcp_rec(AssocList)) :-
	memberchk(uid = _Uid, AssocList),
	!.
rcp_rec(AssocList, rcp_rec(RTerm)) :-
	rcp_rec_uid_store(AssocList, Uid),
	append(AssocList, [uid = Uid], RTerm).

% range (to list)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
rcp_range([Lb .. Ub], List) :-
	!,
	gen_integers(Lb, Ub, List).
rcp_range(List, List) :-
	is_list(List).

% coumpound call
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
rcp_call(Term, R) :-
	Term =.. [F | Args],
	append(Args, [R], FArgs),
	TermToCall =.. [F | FArgs],
	call(TermToCall).

% call
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
rcp_fol_call(Term) :-
	var(Term),
	!,
	Term #= 1.
rcp_fol_call(Term) :-
	integer(Term),
	!,
	Term #= 1.
rcp_fol_call(Term) :-
	call(Term).

% variables (source definition's head)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
variables_bag(Expr) -->
	{
	 var(Expr)
	},
	!,
 	[Expr].
variables_bag(Expr) -->
	{
	 atomic(Expr)
	},
	!.
variables_bag(Expr) -->
	{
	 functor(Expr, _, K)
	},
	variables_bag(0, K, Expr).

variables_bag(K, K, _) -->
	!.
variables_bag(I, K, Expr) -->
	{
	 J is I+1,
	 arg(J, Expr, Arg)
	},
	variables_bag(Arg),
	variables_bag(J, K, Expr).

% variables (source definition's head)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
rcp_variables_bag(Expr) -->
	{
	 var(Expr),
	 rcp_anavar(Expr, N)
	},
	!,
 	[rcp_var(N, Expr)].
rcp_variables_bag(Expr) -->
	{
	 atomic(Expr)
	},
	!.
rcp_variables_bag(Expr) -->
	{
	 functor(Expr, _, K)
	},
	rcp_variables_bag(0, K, Expr).

rcp_variables_bag(K, K, _) -->
	!.
rcp_variables_bag(I, K, Expr) -->
	{
	 J is I+1,
	 arg(J, Expr, Arg)
	},
	rcp_variables_bag(Arg),
	rcp_variables_bag(J, K, Expr).

% multiset
%%%%%%%%%%%%%%%%%%%
variables(Expr, Bag) :-
	variables_bag(Expr, Bag, []).

rcp_variables(Expr, Bag) :-
	rcp_variables_bag(Expr, Bag, []).

% set
%%%%%%%%%%%%%%%%%%%
variables_set(Expr, Set) :-
	variables(Expr, Bag),
	list_to_set(Bag, Set).

rcp_variables_set(Expr, Set) :-
	rcp_variables(Expr, Bag),
	list_to_set(Bag, Set).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% r2cp builtins and FD operators handling
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% arithmetic
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
max(A, B, R) :- call(R #= max(A, B)).

min(A, B, R) :- call(R #= min(A, B)).

log(A, B, R) :- R is ceiling(log(A, B)).

exp(A, B, R) :- R is ceiling(exp(A, B)).

% comparison (reified)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
fd_lt(A, B, R) :- call(R #<=> A #< B).

fd_leq(A, B, R) :- call(R #<=> A #=< B).

fd_eq(A, B, R) :- call(R #<=> A #= B).

fd_geq(A, B, R) :- call(R #<=> A #>= B).

fd_gt(A, B, R) :- call(R #<=> A #> B).

fd_neq(A, B, R) :- call(R #<=> A #\= B).

% comparison (direct)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
fd_lt(A, B) :- call(A #< B).

fd_leq(A, B) :- call(A #=< B).

fd_eq(A, B) :- call(A #= B).

fd_geq(A, B) :- call(A #>= B).

fd_gt(A, B) :- call(A #> B).

fd_neq(A, B) :- call(A #\= B).

% assignment
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
fd_aeq(A, B) :- call(A #= B).

% logic connectives
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
fd_and(P, Q, R) :- R #<=> (P #/\ Q).

fd_not(P, R) :- R #<=> (#\ P).

fd_or(P, Q, R) :- R #<=> (P #\/ Q).

fd_implies(P, Q, R) :- R #<=> (P #=> Q).

fd_equiv(P, Q, R) :- R #<=> (P #<=> Q).

% geost global constraint
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
geost(Objects, ShiftedBoxes, ExternalConstraints) :-
	geosr(Objects,
	      ShiftedBoxes,
	      [(origin(O,D)---> O^x(D)),
	       (size(S,D)--->S^l(D)),
	       (end(O,D)--->
		   origin(O,D)+fold(S,sboxes([O^sid]),+,0,size(S,D))),
	       (overlap_sym(O1,O2,D)--->
		   end(O1,D)#>origin(O2,D)#/\end(O2,D)#>origin(O1,D)),
	       (overlap(O1,O2,Dims)--->
		   forall(D,Dims,overlap_sym(O1,O2,D))),
	       (non_overlapping(Items,Dims)--->
		   forall(O1,Items,
			  forall(O2,Items,
				 O1^oid#<O2^oid#=> #\overlap(O1,O2,Dims))))
	       | ExternalConstraints]).

% integers
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
integer(N, true) :- integer(N).

negative(N, true) :- integer(N), N<0.

positive(N, true) :- integer(N), N>=0.

% list
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
'.'(A, B, R) :- !, R = '.'(A, B).

% logic
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
','(1, 1, 1) :- !.
','(1, B, R) :- !, B, R = 1.
','(A, 1, R) :- !, A, R = 1.
','(A, B, R) :- (A , B), R = 1.

';'(1, 1, 1) :- !.
';'(1, _B, 1) :- !.
';'(_A, 1, 1) :- !.
';'(A, B, R) :- (A ; B), R = 1.

% arithmetic
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
+(A, B, R) :- R #= A + B.

-(A, B, R) :- R #= A - B.

*(A, B, R) :- R #= A * B.

/(A, B, R) :- R #= A / B.

% comparison
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
<(A, B, true) :- A < B.

=<(A, B, true) :- A =< B.

=:=(A, B, true) :- A =:= B.

>=(A, B, true) :- A >= B.

>(A, B, true) :- A > B.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% labeling and heuristics handling
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% variable ordering
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
variable_ordering(CriteriaList) :-
	retractall(rcp_var_ordering(_)),
	assert(rcp_var_ordering(CriteriaList)).	

% value ordering
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
value_ordering(CriteriaList) :-
	retractall(rcp_val_ordering(_)),
	assert(rcp_val_ordering(CriteriaList)).

% top-level labeling
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
rcp_labeling([]) :-
	!.
rcp_labeling(VarList) :-
	rcp_var_ordering(VarCriteriaList),
	rcp_val_ordering(ValCriteriaList),
	rcp_variables(VarList, RcpVarList1),
	delete_bounded_rcp_variables(RcpVarList1, RcpVarList),
	rcp_labeling(RcpVarList, VarCriteriaList, ValCriteriaList).

% heuritics labeling
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
rcp_labeling([], _, _) :-
	!.
rcp_labeling(RcpVarList, VarCriteriaList, ValCriteriaList) :-
				%nl,nl,write('-labeling-\t'),
				%write(RcpVarList),nl,
	select_variable(RcpVarList, VarCriteriaList, BestRcpVar),
				%write('-selectingVar-\t'),
				%BestRcpVar = rcp_var(N, _),
				%write(N),nl,
	assign_variable(BestRcpVar, ValCriteriaList),	
	delete_bounded_rcp_variables(RcpVarList, UpdRcpVarList),
	rcp_labeling(UpdRcpVarList, VarCriteriaList, ValCriteriaList).

assign_variable(rcp_var(N, Var), ValCriteriaList) :-
	select_value_enum(rcp_var(N, Var), ValCriteriaList, ValEnum),
	rcp_var(N, Var),
					%write('-selectingVals-\t'),
					%write(ValEnum),nl,nl,
	labeling(ValEnum, [Var]).

select_value_enum(_, [], [up]) :-
	!.
select_value_enum(rcp_var(from(Name,N,I), Var), [ValCriterion | ValCriteriaList], ValEnum) :-
				%write('-valCriterion-\t'),
				%write(ValCriterion),nl,
				%write(rcp_var(from(Name,N,I), Var)),nl,
	ValCriterion =.. [Order, Term],
	Term =.. [F, NVarOrderCrit, Vars],
	TermToCall =.. [F, NVarOrderCrit, Name, Vars, Result],	
	on_exception(_Error,
		     call(TermToCall),
		     Result = undefined),
	select_value_enum(rcp_var(from(Name,N,I), Var), ValCriteriaList, ValEnum1),
	(
	  Result == Var
	->
	  append(ValEnum1, [Order], ValEnum)
	;
	  ValEnum = ValEnum1
	).

delete_bounded_rcp_variables([], []).
delete_bounded_rcp_variables([rcp_var(N, V) | T],  VT) :-
	rcp_var(N, V),
	integer(V),
	!,
	delete_bounded_rcp_variables(T, VT).
delete_bounded_rcp_variables([V | T], [V | VT]) :-
	delete_bounded_rcp_variables(T, VT).

% variable selection
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
select_variable([BestVar | _], [], BestVar) :-
	!.
select_variable(RcpVarList, VarOrderCriteria, BestVar) :-
	
	crit_order_list(VarOrderCriteria, CritOrderList),
				%write('-CritOrderList-\t\t'),
				%write(CritOrderList),nl,
	var_crit_vectors(RcpVarList, 1, VarOrderCriteria, VarCritValList),
				%write('-VarCritValList-\t\t'),nl,
				%write(VarCritValList),nl,
	order_crit(1, CritOrderList, VarCritValList, [_-(_-BestVar)|_]).

crit_order_list([], []).
crit_order_list([LexVarOrderCrit | LexVarOrderCrits], [Order | OrderList]) :-
	LexVarOrderCrit =.. [Order | _],
	crit_order_list(LexVarOrderCrits, OrderList).

var_crit_vectors([], _, _, []).
var_crit_vectors([rcp_var(Name, Value) | Vars],
		 N,
		 LexVarOrderCriteria,
		 [LexTerm-(N-rcp_var(Name, Value)) | LexTerms]) :-
	var_crit_vector(Name, LexVarOrderCriteria, LexTerm),	
	M is N + 1,
	var_crit_vectors(Vars, M, LexVarOrderCriteria, LexTerms).

var_crit_vector(_, [], []).
var_crit_vector(from(Head, N, I),
		[LexVarOrderCriterion | LexVarOrderCriteria],
		[LexComponent | LexComponents]) :-
	LexVarOrderCriterion =.. [_C, Term],
	Term =.. [F, NVarOrderCrit, Vars],
	TermToCall =.. [F, NVarOrderCrit, Head, Vars, LexComponent],
	on_exception(_Error,
		     call(TermToCall),
		     LexComponent = undefined),
	var_crit_vector(from(Head, N, I), LexVarOrderCriteria, LexComponents).

order_crit(N, OrderList, X, X) :-
	length(OrderList, M),
	N > M,	
	!.
order_crit(NCrit, OrderList, LexValGroup, OrderedLexVals) :-
				%write('-ncrit-\t\t'), write(NCrit),nl,
				%write('-lexico-\t'), write(LexValGroup),nl,
	select_criterion(LexValGroup, NCrit, Lex),
				%write('-lex-\t\t'), write(Lex),nl,
	nth1(NCrit, OrderList, Order),
				%write('-order-\t\t'), write(Order),nl,	
	keyclumps(Order, Lex, [OrderedLexClumped | _]),
				%write('-keyclumps-\t'),write(OrderedLexClumped),nl,
	order_full_lex(OrderedLexClumped, LexValGroup, OrderedFullLex),
				%write('-orderedFullLex-'), write(OrderedFullLex),nl,nl,
	MCrit is NCrit + 1,
	order_crit(MCrit, OrderList, OrderedFullLex, OrderedLexVals).

select_criterion([], _, []).
select_criterion([List-Tag | T], N, [Crit-Tag | Crits]) :-
	nth1(N, List, Crit),
	select_criterion(T, N, Crits).

order_full_lex([], _, []).
order_full_lex([_-T | Comps], LexValGroup, [C-T | FullComps]) :-
	order_full_lex(Comps, LexValGroup, FullComps),
	memberchk(C-T, LexValGroup).

% key clumps
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
keyclumps(is, Lex, OrderedLexClumped) :-
	!,
	keyclumps_is(Lex, OrderedLexClumped).
keyclumps(Order, Lex, OrderedLexClumped) :-
	samsort(Order, Lex, OrderedLex),					
	keyclumps(OrderedLex, OrderedLexClumped).

keyclumps_is(Lex, OrderedLexClumped) :-
	keyclumps_is_(Lex, OrderedLexYes, OrderedLexNo),
	(
	  OrderedLexYes == []
	->
	  OrderedLexClumped = [OrderedLexNo]
	 ;
	  OrderedLexClumped = [OrderedLexYes, OrderedLexNo]
	).

keyclumps_is_([], [], []) :-
	!.
keyclumps_is_([W-(N-V) | T], Y, Nos) :-
	var(W),
	rcp_anavar(W, WName),
	rcp_var(WName, W) == V,
	!,
	keyclumps_is_(T, Yess, Nos),
	append([V-(N-V)], Yess, Y).
keyclumps_is_([H | T], Yess, N) :-
	keyclumps_is_(T, Yess, Nos),
	append([H], Nos,  N).	

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% term comparison
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% greatest (variable ordering)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
greatest(_-_, undefined-_) :-
	!,
	fail.
greatest(undefined-X, _-Y) :-
	!,
	X @=< Y.	
greatest(T1-X, T2-Y) :-
	X @=< Y,
	T1 @>= T2.

% least (variable ordering)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
least(undefined-_, _-_) :-
	!,
	fail.
least(_-X, undefined-Y) :-
	!,
	X @=< Y.
least(T1-X, T2-Y) :-
	X @=< Y,
	T1 @=< T2.

% any (variable ordering)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
any(_-X, _-Y) :-
	X @=< Y.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% utilities
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% generate integer list
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
gen_integers(L, U, Ints) :- 
	gen_integers(L, U, L, Ints).

gen_integers(L, U, _, []) :-
	L > U,
	!.
gen_integers(_, U, U, [U]) :-
	!.
gen_integers(L, U, I, [I | Is]) :-
	N is I + 1,
	N =< U,
	N >= L,
	gen_integers(L, U, N, Is).

% copy_term (preserves all variables but V)
%%%%%%%%%%%%%%%%%%%%
copy_term_var(T, V, U) :-
	copy_term_var(T, V, _, _, U).

copy_term_var(T, V, Assoc, NewAssoc, U) :-
	var(T),
	T == V,
	!,
	(
	  empty_avl(Assoc)
	->
	  avl_store(V, Assoc, U, NewAssoc)
	;
	  avl_fetch(V, Assoc, U),
	  NewAssoc = Assoc
	).
copy_term_var(T, _, Assoc, Assoc, T) :-
	var(T),
	!.
copy_term_var(T, V, Assoc, NewAssoc, U) :-
	is_list(T),
	!,
	copy_term_var_list(T, V, Assoc, NewAssoc, U).
copy_term_var(T, V, Assoc, NewAssoc, U) :-
	compound(T),
	!,
	T =.. [F | Args],
	copy_term_var_list(Args, V, Assoc, NewAssoc, ArgsCopy),
	U =.. [F | ArgsCopy].
copy_term_var(T, _, Assoc, Assoc, T).
 
copy_term_var_list([], _V, Assoc, Assoc, []).
copy_term_var_list([T|Ts], V, Assoc, NewAssoc, [U|Us]) :-
	copy_term_var(T, V, Assoc, NewAssoc1, U),
	copy_term_var_list(Ts, V, NewAssoc1, NewAssoc, Us).

% variable instantiation 
%%%%%%%%%%%%%%%%%%%%
instantiate_variable(Var, Elem, Expr, InstExpr) :-
	copy_term_var(Var-Expr, Var, InstVar-InstExpr),
	InstVar = Elem.

% record uids handling
%%%%%%%%%%%%%%%%%%%%
init_rcp_rec_uid :-
	recorded(rcp_rec_uid(Record, Uid), rcp_rec_uid(Record, Uid), Ref),
	!,
	erase(Ref),
	init_rcp_rec_uid.
init_rcp_rec_uid :-
	counter_init(rcp_rec_uid_cpt).

rcp_rec_uid_store(Record0, Uid) :-
	rcp_name_vars_in_rec(Record0, Record),
	recorded(rcp_rec_uid(Record, Uid), rcp_rec_uid(Record, Uid), _),
	!.
rcp_rec_uid_store(Record0, Uid) :-	
	counter_inc(rcp_rec_uid_cpt, Uid),
	rcp_name_vars_in_rec(Record0, Record),
	recordz(rcp_rec_uid(Record, Uid), rcp_rec_uid(Record, Uid), _).

rcp_name_vars_in_rec([], []) :-
	!.
rcp_name_vars_in_rec([K = V0 | Rec0], [K = V | Rec]) :-
	var(V0),
	!,
	rcp_anavar(V0, V),
	rcp_name_vars_in_rec(Rec0, Rec).
rcp_name_vars_in_rec([K = V | Rec0], [K = V | Rec]) :-
	rcp_name_vars_in_rec(Rec0, Rec).

% counters handling
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
:- dynamic rcp_counter/2.

counter_init(Counter) :-
	retractall(rcp_counter(Counter,_)).

counter_inc(Counter, N) :-	
	retract(rcp_counter(Counter, N)),
	!,
	M is N + 1,
	assert(rcp_counter(Counter, M)).
counter_inc(Counter, N) :-
	N is 1,
	assert(rcp_counter(Counter, 2)).

% flips Term arguments
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
flip_list([], []).
flip_list([X], [X]).
flip_list([X1, X2 | Xs], [X2, X1 | Ys]) :-
	flip_list(Xs, Ys).

flip(Term, FTerm) :- 
	is_list(Term),
	!,
	flip_list(Term, FTerm).
flip(Term, FTerm) :-
	compound(Term),
	!,
	Term =.. [Op | Args],
	flip_list(Args,FArgs),
	FTerm =.. [Op | FArgs].
flip(Term, Term).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% higher-order functions (iterators)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% % ho macros
% %%%%%%%%%%%%%%%%%%%%
% forall(X, L, G, R) :-
% 				%write(forall(X,L,G)),nl,
% 	rcp_range(L, List),
% 	fold_left(X, List, ',', true, G, R).

% forall(X, L, G) :-
% 				%write(forall(X,L,G)),nl,
% 	rcp_range(L, List),
% 	fold_left(X, List, ',', true, G, _).

% exists(X, L, G, R) :-
% 				%write(exists(X,L,G)),nl,
% 	rcp_range(L, List),
% 	fold_left(X, List, ';', fail, G, R).

% exists(X, L, G) :-
% 				%write(exists(X,L,G)),nl,
% 	rcp_range(L, List),
% 	fold_left(X, List, ';', fail, G, _).

% let(X, E, G) :-
% 				%write(let(X,[E],G)),nl,
% 	E =.. [F|Args],
% 	append(Args, [R], RArgs),
% 	D =.. [F | RArgs],
% 	call(D),
% 	let(X, R, G, _).

% map(X, L, G, R) :-
% 				%write(map(X,L,G)),nl,
% 	rcp_range(L, List),
% 	fold_right(X, List, '.', [], G, R).

% % let
% %%%%%%%%%%%%%%%%%%%%
% let(V, X, G, R) :-
% 				%write(let(V, X, G)),nl,
% 	(
% 	  (
% 	    is_list(X)
% 	  ;
% 	    number(X)
% 	  )
% 	->
% 	  RX = X
% 	;
% 	  X =.. [F | XArgs],
% 	  append(XArgs, [RX], RXArgs),
% 	  Y =.. [F | RXArgs],
% 	  call(Y)
% 	),
% 	instantiate_variable(V, RX, G, Gc),
% 	Gc =.. [_ | Args],
% 	last(Args, R),
% 	call(Gc).

% % fold right
% %%%%%%%%%%%%%%%%%%%%
% foldr(X, L, F, I, G, R) :-
% 				%write(foldr(X, L, F, I, G)),nl,
% 	rcp_range(L, List),
% 	fold_right(X, List, F, I, G, R).

% fold_right(_, [],  _, I, _, I) :-
% 	!.
% fold_right(V, [X | Xs], F, I, G, R) :-
% 				%write(fold_right(V, X)),nl,
% 	fold_right(V, Xs, F, I, G, R1),
% 	instantiate_variable(V, X, G, Gc),	
% 	(
% 	  number(Gc)
% 	->
% 	  RGc = Gc
% 	;
% 	  Gc =.. [_H | GcArgs],
% 	  last(GcArgs, RGc),
% 	  instantiate_variable(RGc, RGc1, Gc, Gc1),
% 	  call(Gc1)
% 	),
% 	T =.. [F, RGc1, R1, R],
% 	call(T).

% % fold left
% %%%%%%%%%%%%%%%%%%%%
% foldl(X, L, F, I, G, R) :-
% 				%write(foldl(X, L, F, I, G)),nl,
% 	rcp_range(L, List),
% 	fold_left(X, List, F, I, G, R).

% fold_left(_, [], _, I, _, I) :-
% 	!.
% fold_left(V, [X | Xs], F, I, G, R) :-
% 				%write(fold_left(V, X)),nl,nl,	
% 	instantiate_variable(V, X, G, Gc),				
% 	(
% 	  number(Gc)
% 	->
% 	  RGc = Gc
% 	;
% 	  Gc =.. [_ | GcArgs],
% 	  last(GcArgs, RGc),
% 	  instantiate_variable(RGc, RGc1, Gc, Gc1),
% 	  call(Gc1)
% 	),	    
% 	T =.. [F, I, RGc1, R1],
% 	call(T),
% 	fold_left(V, Xs, F, R1, G, R).

% % loop until TG[V/X] is true,
% % G[V/X] otherwise
% %%%%%%%%%%%%%%%%%%%%
% until(V, X, _, TG) :-
% 	let(V, X, TG, _),
% 	!.
% until(V, X, G, TG) :-
% 	let(V, X, G, R),
% 	(
% 	  let(W,R,negative(W),_)
% 	->
% 	  true
% 	;
% 	  until(V, R, G, TG)
% 	).

% memberchk of assoc. values
% on term equality, not unification
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% mymemberchk(_, []) :-
% 	!.
% mymemberchk((Name = Value), [Name0 = Value0 | _]) :-
% 	Value == Value0,
% 	!,
% 	Name = Name0.
% mymemberchk((Name = Value), [_ = _ | Tail]) :-
% 	mymemberchk((Name = Value), Tail).

% rcp_anavar(Value, Name) :-
% 	execution_state(break_level(0), private(Priv)),
% 	memberchk(r2cp(Priv0), Priv),
% 	mymemberchk((Name = Value), Priv0).

/*

bin(Bin), items(Items), Dims=[1,2,3], psa_bin_packing(Bin, Items, Dims, 1), rcp_var(from(o2,0,1), O2),
	rcp_var(from(o2,0,2), O2_2),rcp_var(from(o2,0,3), O2_3),rcp_var(from(o3,0,1), O3),
	rcp_var(from(o3,0,2), O3_2),rcp_var(from(o3,0,3), O3_3),rcp_var(from(o4,0,1), O4),
	rcp_var(from(o4,0,2), O4_2),rcp_var(from(o4,0,3), O4_3),	rcp_var(from(o5,0,1), O5),
	rcp_var(from(o5,0,2), O5_2),rcp_var(from(o5,0,3), O5_3),	rcp_var(from(o6,0,1), O6),
	rcp_var(from(o6,0,2), O6_2),rcp_var(from(o6,0,3), O6_3),	rcp_var(from(o7,0,1), O7),
	rcp_var(from(o7,0,2), O7_2),rcp_var(from(o7,0,3), O7_3),	rcp_var(from(o8,0,1), O8),
	rcp_var(from(o8,0,2), O8_2),rcp_var(from(o8,0,3), O8_3),
	rcp_var(from(o9,0,1), O9),rcp_var(from(o9,0,2), O9_2),	rcp_var(from(o9,0,3), O9_3),
	rcp_var(from(o10,0,1), O10),	rcp_var(from(o10,0,2), O10_2),	rcp_var(from(o10,0,3), O10_3).


	tasks_domain(1),
	precedences(1),
	distances(1),
	minimize((disjunctives_search(1), labeling([], [Last])), Last).

*/

/*

interval_split_foldl_1_search([], _, 1) :- 
	!.
interval_split_foldl_1_search([Cut_380|Tail], [L0,Max0,Min0,X0], _R_) :- 
	_A_=[X0,L0],
	(
	  var(X0)
	->
	  rcp_anavar(X0, XName0),
	  XName0 = from(item(rcp_rec([boxes=[rcp_rec([size=[XName,XName],uid=_])],positions=[[_,_]]])),_,_)
	;
	  XName = X0-ground
	),
	nth1(1, _A_, X),
	nth1(2, _A_, L),
	fd_min(X, XMin),
	XDLBL is XMin + L,
	(
	  write(leq(XName, XDLBL)),nl,
	  
	  fd_leq(X, XDLBL, _Leq_),
	  rcp_fol_call(_Leq_)
	;
	  write(gt(XName, XDLBL)),nl,
	  
	  fd_gt(X, XDLBL, _Gt_),
	  rcp_fol_call(_Gt_),		
	  interval_split_foldl_1_search(Tail, [L0,Max0,Min0,X0], _R_)
	).

--

statistics(runtime, [T0| _]),

write(rect_try(Bin_3,Bin,Bin_2)),nl,

write('interval(Xs)'),nl,

write('interval(Ys)'),nl,

,
statistics(runtime, [T1|_]),
T is T1 - T0,
format('took ~3d sec.~n', [T])


N=23
took 4101.430 sec.
X = [4352,68,64] 
*/

